<?php

namespace Drupal\component_builder;

use Drupal\component_builder\Entity\ComponentItemTypeInterface;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\PhpStorage\PhpStorageFactory;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\taxonomy\Entity\Term;
use Drupal\taxonomy\Entity\Vocabulary;
use Symfony\Component\Yaml\Yaml;

/**
 * The ImportConfigComponent service.
 */
class ImportConfigComponent {

  CONST COMPONENT_EXCEPT = ['table_column'];

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The module handler to invoke the alter hook.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The module.
   *
   * @var string
   */
  protected $module;

  /**
   * The image style entity storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $imageStyleStorage;

  /**
   * ImportConfigComponent constructor.
   *
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   */
  public function __construct(EntityFieldManagerInterface $entity_field_manager = NULL, EntityTypeManagerInterface $entity_type_manager = NULL, ModuleHandlerInterface $module_handler = NULL) {
    $this->entityFieldManager = \Drupal::service('entity_field.manager');
    $this->moduleHandler = $module_handler ?: \Drupal::moduleHandler();
    $this->entityTypeManager = $entity_type_manager ?: \Drupal::entityTypeManager();
    $this->module = 'component_builder';
    $this->imageStyleStorage = $this->entityTypeManager->getStorage('image_style');
  }

  /**
   * {@inheritdoc}
   */
  public function createComponentItem(string $bundle, string $label, string $module = 'component_builder') {
    $this->module = $module;

    $componentItemTypeStorage = $this->entityTypeManager->getStorage('component_item_type');
    if ($componentItemTypeStorage) {
      $componentType = $this->entityTypeManager->getStorage('component_item_type')
        ->load($bundle);
      if (!$componentType) {
        if ($bundle === 'composite') {
          $this->createTerm($label);
          return TRUE;
        }
        else {
          $componentName = get_component_name($bundle);
          $componentPath = $this->getPathComponentItem($componentName);
          $pathConfig = $componentPath . "/config/component_builder.component_item_type.$bundle.yml";
          $values = $this->parseYaml($pathConfig);
          if ($values) {
            $newComponentType = $componentItemTypeStorage->create($values);
            $newComponentType->save();
            $userRoleStorage = $this->entityTypeManager->getStorage('user_role');
            $role = $userRoleStorage->load('administrator');
            $issueAccessPermissions = [
              "create $bundle content",
              "delete any $bundle content",
              "delete own $bundle content",
              "edit any $bundle content",
              "edit own $bundle content",
            ];
            foreach ($issueAccessPermissions as $permission) {
              $role->grantPermission($permission);
            }
            $role->save();
            // $this->generatePathautoPattern('node', $machine_name);
            $this->createFieldsComponent('component_item', $bundle);
            $this->createImageStylesComponent($bundle);
            $this->createDisplay('component_item', $bundle);
            if (!in_array($bundle, self::COMPONENT_EXCEPT)) {
              $this->createTerm($label);
            }

            // Create component wrapper.
            $this->createComponentWrapperField($newComponentType);
            return $bundle;
          }
        }
      }
      else {
        return NULL;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createFieldsComponent($entity = '', $bundle = ''): void {
    $componentName = get_component_name($bundle);
    $componentPath = $this->getPathComponentItem($componentName);
    $configPath = $componentPath . '/config/config.yml';
    $yml = $this->parseYaml($configPath);
    if ($yml && isset($yml['fields'])) {
      $fieldDefines = $this->entityFieldManager->getFieldDefinitions($entity, $bundle);
      $fields = $yml['fields'];
      foreach ($fields as $fieldName => $label) {
        if (!array_key_exists($fieldName, $fieldDefines)) {
          $this->createField($fieldName, $entity, $bundle, $componentPath);
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createDisplay($entity, $bundle): void {
    $componentName = get_component_name($bundle);
    $componentPath = $this->getPathComponentItem($componentName);
    $this->setFormDisplay($componentPath, $entity, $bundle, 'default');
    $this->setViewDisplay($componentPath, $entity, $bundle, 'default');
  }

  /**
   * {@inheritdoc}
   */
  public function createField($fieldName = '', $entity = '', $bundle = '', $componentPath = ''): void {
    $fieldStorage = FieldStorageConfig::loadByName($entity, $fieldName);
    if (!$fieldStorage) {
      $pathStorageConfig = $componentPath . "/config/field.storage.component_item.$fieldName.yml";
      $this->importFieldStorageYmlConfig($pathStorageConfig);
    }
    $pathConfig = $componentPath . "/config/field.field.component_item.$bundle.$fieldName.yml";
    $this->importFieldYmlConfig($pathConfig);
  }

  /**
   * {@inheritdoc}
   */
  public function setFormDisplay($componentPath, $entityType, $bundle, $form_mode) {
    $entityFormDisplayStorage = $this->entityTypeManager->getStorage('entity_form_display');
    $entityFormDisplay = $entityFormDisplayStorage->load("$entityType.$bundle.$form_mode");
    if (!$entityFormDisplay) {
      $pathConfig = $componentPath . "/config/core.entity_form_display.$entityType.$bundle.$form_mode.yml";
      $values = $this->parseYaml($pathConfig);
      if ($values) {
        $entity_form_display = $entityFormDisplayStorage->create($values);
        $entity_form_display->save();
      }
    }
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function setViewDisplay($componentPath, $entityType, $bundle, $form_mode) {
    $entityFormDisplayStorage = $this->entityTypeManager->getStorage('entity_view_display');
    $entityFormDisplay = $entityFormDisplayStorage->load("$entityType.$bundle.$form_mode");
    if (!$entityFormDisplay) {
      $pathConfig = $componentPath . "/config/core.entity_view_display.$entityType.$bundle.$form_mode.yml";
      $values = $this->parseYaml($pathConfig);
      if ($values) {
        $entity_form_display = $entityFormDisplayStorage->create($values);
        $entity_form_display->save();
      }
    }
    return TRUE;
  }

  /**
   * Flush all cache.
   */
  public function flushAll() {
    drupal_flush_all_caches();
  }

  /**
   * Clear cache bin.
   *
   * @param string $cache_name
   *   Cache name.
   */
  public function flush($cache_name = '') {
    if (empty($cache_name)) {
      $this->flushAll();
    }
    else {
      foreach (Cache::getBins() as $service_id => $cache_backend) {
        if (is_array($cache_name)) {
          if (in_array($service_id, $cache_name)) {
            $cache_backend->deleteAll();
          }
        }
        else {
          if ($cache_name == 'twig') {
            PhpStorageFactory::get('twig')->deleteAll();
          }
          elseif ($service_id == $cache_name) {
            $cache_backend->deleteAll();
          }
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function importFieldYmlConfig($path_to_file): void {
    $field_string = file_get_contents($path_to_file);
    $field_config = Yaml::parse($field_string);
    if ($field_config) {
      $fieldConfigStorage = $this->entityTypeManager->getStorage('field_config');
      $fieldStorage = $fieldConfigStorage->create($field_config);
      $fieldStorage->save();
      $this->flush('twig');
    }
  }

  public function createTerm(string $label) {
    $vocabulary = $this->entityTypeManager->getStorage('taxonomy_vocabulary')->load('component_types');
    if (!$vocabulary) {
      $this->createVocabulary('Component Types');
    }
    $term = $this->entityTypeManager->getStorage('taxonomy_term')->getQuery()
      ->condition('name', $label)
      ->condition('vid', 'component_types')
      ->execute();
    if (!$term) {
      $term = Term::create([
        'name' => $label,
        'description' => '',
        'parent' => [0],
        'vid' => 'component_types',
      ]);
      $term->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createVocabulary(string $label) {
    $machine = preg_replace('@[^a-z0-9-]+@', '_', strtolower($label));
    $vocabulary = Vocabulary::create([
      'name' => $label,
      'description' => $label,
      'vid' => $machine,
    ]);
    $vocabulary->save();
  }

  /**
   * {@inheritdoc}
   */
  public function getPathComponentItem(string $componentName): string {
    $module_path = $this->moduleHandler->getModule($this->module)
      ->getPath();
    return $module_path . '/components/' . $componentName;
  }

  /**
   * {@inheritdoc}
   */
  public function parseYaml($path): array {
    if (!file_exists($path)) {
      return [];
    }
    return Yaml::parse(file_get_contents($path)) ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function importFieldStorageYmlConfig($path) {
    $fieldConfig = $this->parseYaml($path);
    if (!empty($fieldConfig)) {
      $storage = $this->entityTypeManager->getStorage('field_storage_config');
      $fieldStorage = $storage->create($fieldConfig);
      $fieldStorage->save();
      $this->flush('twig');
    }
    return $fieldStorage;
  }

  /**
   * {@inheritdoc}
   */
  public function createComponentWrapperField(ComponentItemTypeInterface $componentItemType): void {
    $fieldName = 'field_' . $componentItemType->id();
    $fieldDefinitions = $this->entityFieldManager->getFieldDefinitions('component_wrapper', 'component_wrapper');

    if (!in_array($fieldName, array_keys($fieldDefinitions))) {
      if (!FieldStorageConfig::loadByName('component_wrapper', $fieldName)) {
        FieldStorageConfig::create([
          'field_name' => $fieldName,
          'type' => 'entity_reference',
          'entity_type' => 'component_wrapper',
          'cardinality' => FieldStorageDefinitionInterface::CARDINALITY_UNLIMITED,
          'settings' => [
            'target_type' => 'component_item',
          ],
        ])->save();
      }
      if (!FieldConfig::loadByName('component_wrapper', 'component_wrapper', $fieldName)) {
        FieldConfig::create([
          'field_name' => $fieldName,
          'entity_type' => 'component_wrapper',
          'bundle' => 'component_wrapper',
          'label' => $componentItemType->label() . ' Items',
          'settings' => [
            'handler' => 'default',
            'handler_settings' => [
              'target_bundles' => [$componentItemType->id()],
            ],
          ],
        ])->save();
      }
      $this->entityFieldManager->clearCachedFieldDefinitions();

      /** @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface $display_repository */
      $display_repository = \Drupal::service('entity_display.repository');

      $formDisplay = $display_repository->getFormDisplay('component_wrapper', 'component_wrapper');
      $formDisplay->setComponent($fieldName, [
        'type' => 'inline_entity_form_complex',
        'weight' => 5,
        'settings' => [
          'override_labels' => TRUE,
          'label_singular' => t('Item'),
          'label_plural' => t('Items'),
        ],
      ])->save();
      $viewDisplay = $display_repository->getViewDisplay('component_wrapper', 'component_wrapper');
      $viewDisplay->setComponent($fieldName, [
        'type' => 'entity_reference_entity_view',
        'label' => 'hidden',
        'settings' => [
          'view_mode' => 'full',
        ],
      ])->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createImageStyle(string $name, string $label, array $effect_info): void {
    $style = $this->imageStyleStorage->load($name);
    if (!$style) {
      $style = $this->imageStyleStorage->create(array('name' => $name, 'label' => $label));
      $effect = \Drupal::service('plugin.manager.image.effect')->createInstance($effect_info['id'], $effect_info);
      $style->addImageEffect($effect->getConfiguration());
      $style->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createImageStylesComponent($bundle = ''): void {
    $componentName = get_component_name($bundle);
    $componentPath = $this->getPathComponentItem($componentName);
    $configPath = $componentPath . '/config/config.yml';
    $yml = $this->parseYaml($configPath);
    if ($yml && isset($yml['image_styles'])) {
      $imageStylesDefinitions = \Drupal::entityTypeManager()->getStorage('image_style')->getQuery()->execute();
      $imageStyles = $yml['image_styles'];
      foreach ($imageStyles as  $imageStyle) {
        $image_style_path = $componentPath . "/config/image.style.$imageStyle.yml";
        $yml = $this->parseYaml($image_style_path);
        if ($yml) {
          $name = $yml['name'];
          if ($name && !isset($imageStylesDefinitions[$name])) {
            $label = $yml['label'];
            $effect_info = $yml['effects'];
            if ($effect_info) {
              $effect_info = current($effect_info);
              if (isset($effect_info['uuid'])) {
                unset($effect_info['uuid']);
              }
            }
            $this->createImageStyle($name, $label, $effect_info);
          }
        }
      }
    }
  }

}
